function [D,Gamma,err,gerr] = ksvd(params,verbose,msgdelta)
%KSVD K-SVD dictionary training.
%  [D,GAMMA] = KSVD(PARAMS) runs the K-SVD dictionary training algorithm on
%  the specified set of signals, returning the trained dictionary D and the
%  signal representations matrix GAMMA.
%
%  KSVD has two modes of operation: sparsity-constrained and
%  error-constrained. For sparsity-constrained minimization, the
%  optimization problem is given by
%
%      min  |X-D*GAMMA|_F^2      s.t.  |Gamma_i|_0 <= T
%    D,Gamma
%
%  where X is the set of training signals, Gamma_i is the i-th column of
%  Gamma, and T is the target sparsity. For error-constrained minimization,
%  the optimization problem is given by
%
%      min  |Gamma|_0      s.t.  |X_i - D*Gamma_i|_2 <= EPSILON
%    D,Gamma
%
%  where X_i is the i-th training signal, and EPSILON is the target error.
%
%  [D,GAMMA,ERR] = KSVD(PARAMS) also returns the target function values
%  after each algorithm iteration. For sparsity-constrained minimization,
%  the returned values are
%
%    ERR(D,GAMMA) = RMSE(X,D*GAMMA) = sqrt( |X-D*GAMMA|_F^2 / numel(X) ) .
%
%  For error-constrained minimization, the returned values are
%
%    ERR(D,GAMMA) = mean{ |Gamma_i|_0 } = |Gamma|_0 / size(X,2) .
%
%  Error computation slightly increases function runtime.
%
%  [D,GAMMA,ERR,GERR] = KSVD(PARAMS) computes the target function values on
%  the specified set of test signals as well, usually for the purpose of
%  testing the generalization capability of the dictionary. This requires
%  that the field 'testdata' be present in PARAMS (see below). The length
%  of ERR and GERR is identical.
%
%  [...] = KSVD(PARAMS,VERBOSE) specifies whether the function should print
%  messages as it runs, and their level of detail. When VERBOSE is 1, an
%  iteration counter is displayed each iteration. When VERBOSE is 2, the
%  number of replaced atoms is printed as well. When VERBOSE is 3, the
%  function prints the target function value at each iteration, as well as
%  the test data representation error, if the 'testdata' field is present.
%  When VERBOSE is 4, both the number of replaced atoms as well as the
%  values of the target function are displayed. If VERBOSE is any other
%  value, or is not specified, no information is displayed.
%
%  [...] = KSVD(PARAMS,VERBOSE,MSGDELTA) specifies that the function should
%  print additional status messages within each iteration. A message is
%  printed every MSGDELTA seconds. When MSGDELTA is negative or zero (the
%  default), no such messages are displayed.
%
%
%  Required fields in PARAMS:
%    data        -  the training signals (column vectors).
%    dictsize    -  the number of dictionary elements to train.
%    Tdata/Edata -  number of coefficients (Tdata) or target error (Edata)
%                   when coding each signal. if only one is present, that
%                   value is used. if both are present, Tdata is used
%                   unless the field 'codemode' is specified (see below).
%
%
%  Optional fields in PARAMS:
%    initdict    -  if present, specifies the initial dictionary for the
%                   training process. may either be a matrix (of size
%                   SIZE(data,1) X dictsize), or an array of length
%                   dictsize specifying the indices of the training signals
%                   to be used as the initial atoms. when not present, D is
%                   initialized with randomly selected training signals.
%    codemode    -  specifies whether sparsity-based or error-based coding
%                   should be used in the training. this is useful when
%                   both the Tdata and the Edata field are present in
%                   PARAMS. codemode can be either 'sparsity' or 'error'.
%                   if it is not specified, and both fields are present,
%                   sparsity-based coding takes place.
%    testdata    -  if present, generalization error is computed on this
%                   data.
%
%
%  Optional fields in PARAMS (advanced):
%    iternum     -  number of iterations to perform (default = 20).
%    memusage    -  this parameter can be used to control the memory
%                   requirements of the function. memusage can be either
%                   'high', 'normal' (default value) or 'low'. when set to
%                   'high', both the matrices G=D'*D and DtX=D'*X are
%                   precomputed prior to performing the OMP coding step,
%                   thus increasing both speed and memory requirements of
%                   the method. when set to 'normal', only the matrix G is
%                   precomputed, and when set to 'low', none of the
%                   matrices are precomputed. The 'low' setting is suitable
%                   when the dictionary is highly redundant and memory
%                   resources are low. however, with this setting on,
%                   runtime will increase substantially. see function OMP
%                   for more details.
%    exact       -  specifies whether to update each dictionary atom using
%                   the exact svd computation, or a much quicker
%                   approximation. by default, the quicker computation is
%                   used. however, if this parameter is specified and is
%                   non-zero, the complete svd computation is used instead.
%                   this significantly slowes down the method, but provides
%                   slightly improved results.
%    maxatoms    -  if specified, when error-based coding is used, limits
%                   the number of selected atoms to this value (see
%                   parameter 'maxatoms' in OMP2/NNOMP2). maxatoms has no
%                   effect when sparsity-based coding is used.
%    muthresh    -  this parameter can be used to control the target mutual
%                   incoherence of the trained dictionary, and is typically
%                   between 0.9 and 1. by default, the trained dictionary
%                   is "cleaned" each iteration by eliminating atoms with
%                   correlation > 0.99. setting muthresh to a different
%                   value modifies this threshold to the specified value.
%                   note that muthresh does not impose a strict constraint
%                   on the dictionary, and the dictionary may ultimately
%                   have a larger mutual incoherence. a method to track
%                   this is using the VERBOSE parameter to print the number
%                   of replaced atoms each iteration. when this number
%                   drops near zero, it is more likely that the mutual
%                   incoherence of the dictionary is below muthresh.
%                   setting muthresh to 1 or higher cancels this type of
%                   cleaning completely. muthresh cannot be negative.
%    filename    -  if present, saves the workspace to the file
%                   'filenameN.mat', where N is the itaration number, every 
%                   savenum iterations, as well as at the end of the
%                   algorithm.
%    savenum     -  specifies the number of iterations between file saves
%                   (default=10). A value of 0 is used to specify that a
%                   save is required only at the end of the algorithm and
%                   not between iterations.
%
%
%  References:
%  [1] M. Aharon, M. Elad, and A.M. Bruckstein, "The K-SVD: An Algorithm
%      for Designing of Overcomplete Dictionaries for Sparse
%      Representation", the IEEE Trans. On Signal Processing, Vol. 54, no.
%      11, pp. 4311-4322, November 2006.
%  [2] M. Elad, R. Rubinstein, and M. Zibulevsky, "Efficient Implementation
%      of the K-SVD Algorithm and the Batch-OMP Method", Technical Report -
%      CS, Technion, April 2008.
%
%  See also KSVDDENOISE, OMPDENOISE, OMP, OMP2.


%  Ron Rubinstein
%  Computer Science Department
%  Technion, Haifa 32000 Israel
%  ronrubin@cs
%
%  December 2007


global CODE_SPARSITY CODE_ERROR codemode
global MEM_LOW MEM_NORMAL MEM_HIGH memusage
global nonneg ompfunc ompparams exactsvd

CODE_SPARSITY = 1;
CODE_ERROR = 2;

MEM_LOW = 1;
MEM_NORMAL = 2;
MEM_HIGH = 3;


%%%%% parse input parameters %%%%%


dictsize = params.dictsize;
data = params.data;

ompparams = {};

% coding mode %

if (isfield(params,'codemode'))
  switch lower(params.codemode)
    case 'sparsity'
      codemode = CODE_SPARSITY;
      thresh = params.Tdata;
      XtX = [];
    case 'error'
      codemode = CODE_ERROR;
      thresh = params.Edata;
      XtX = sum(data.*data);
    otherwise
      error('Invalid coding mode specified');
  end
elseif (isfield(params,'Tdata'))
  codemode = CODE_SPARSITY;
  thresh = params.Tdata;
  XtX = [];
elseif (isfield(params,'Edata'))
  codemode = CODE_ERROR;
  thresh = params.Edata;
  XtX = sum(data.*data);
else
  error('Data sparse-coding target not specified');
end


% max number of atoms %

if (codemode==CODE_ERROR && isfield(params,'maxatoms'))
  ompparams{end+1} = 'maxatoms';
  ompparams{end+1} = params.maxatoms;
end


% memory usage %

if (isfield(params,'memusage'))
  switch lower(params.memusage)
    case 'low'
      memusage = MEM_LOW;
    case 'normal'
      memusage = MEM_NORMAL;
    case 'high'
      memusage = MEM_HIGH;
    otherwise
      error('Invalid memory usage mode');
  end
else
  memusage = MEM_NORMAL;
end


% non-negativity %

nonneg = 0;


% iteration count %

if (isfield(params,'iternum'))
  iternum = params.iternum;
else
  iternum = 20;
end


% data saving %

if (isfield(params,'filename'))
  filename = params.filename;
  savenum = 10;
  if (isfield(params,'savenum'))
    savenum = params.savenum;
  end
else
  savenum = -1;
end


% omp function %

if (codemode == CODE_SPARSITY)
  ompfunc = @omp;
else
  ompfunc = @omp2;
end


% verbose mode %

printiter = 0;
printreplaced = 0;
printerr = 0;
printgerr = 0;

if (nargin>=2)
  
  if (verbose>=1 && verbose<=4)
    printiter = 1;
  end
  
  if (verbose==2 || verbose==4)
    printreplaced = 1;
  end
  
  if (verbose==3 || verbose==4)
    printerr = 1;
    if (isfield(params,'testdata'))
      printgerr = 1;
    end
  end
  
  if (nargin>=3)
    if (msgdelta<0), msgdelta=0; end
    ompparams{end+1} = 'messages';
    ompparams{end+1} = msgdelta;
  else
    msgdelta = 0;
  end
  
end


% compute error flag %

comperr = (nargout>=3 || printerr);


% test generalization flag %

testgen = 0;
if (isfield(params,'testdata'))
  testdata = params.testdata;
  if (codemode==CODE_ERROR)
    XtXg = sum(testdata.*testdata);
  else
    XtXg = [];
  end
  if (nargout>=4 || printgerr)
    testgen = 1;
  end
end


% mutual incoherence limit %

if (isfield(params,'muthresh'))
  muthresh = params.muthresh;
else
  muthresh = 0.99;
end
if (muthresh < 0)
  error('invalid muthresh value, must be non-negative');
end


% exact svd computation %

exactsvd = 0;
if (isfield(params,'exact') && params.exact~=0)
  exactsvd = 1;
end


%%%%% initialize the dictionary %%%%%

if (size(data,2) < dictsize)
  error('Number of training signals is smaller than requested dictionary size');
end

if (isfield(params,'initdict'))
  if (ndims(params.initdict)<=2 && any(size(params.initdict)==1) && numel(params.initdict) == dictsize)
    D = data(:,params.initdict(:));
  else
    D = params.initdict(:,1:dictsize);
  end
else
  data_ids = find(colnorms(data)>1e-2);   % ensure no zero data elements are chosen
  perm = randperm(length(data_ids));
  D = data(:,data_ids(perm(1:dictsize)));
end


%%%%% normalize the dictionary %%%%%

D = normcols(D);

err = zeros(1,iternum);
gerr = zeros(1,iternum);

if (codemode == CODE_SPARSITY)
  errstr = 'RMSE';
else
  errstr = 'mean atomnum';
end



%%%%%%%%%%%%%%%%%  main loop  %%%%%%%%%%%%%%%%%


for iter = 1:iternum
  
  G = [];
  if (memusage >= MEM_NORMAL)
    G = D'*D;
  end
  
  
  %%%%%  sparse coding  %%%%%
  
  Gamma = sparsecode(data,D,XtX,G,thresh);
  
  
  %%%%%  dictionary update  %%%%%
  
  replaced_atoms = zeros(1,dictsize);  % mark each atom replaced by optimize_atom
  
  unused_sigs = 1:size(data,2);  % tracks the signals that were used to replace "dead" atoms.
                                 % makes sure the same signal is not selected twice
                                 
  GammaT = Gamma';   % updates of sparse matrices are much quicker for columns
                     % rather than rows, so process is done on Gamma ranspose
  
  p = randperm(dictsize); 
  tic; starttime = toc; currtime = starttime;
  for j = 1:dictsize
    [D(:,p(j)),GammaT(:,p(j)),unused_sigs,replaced_atoms] = optimize_atom(data,D,p(j),GammaT,unused_sigs,replaced_atoms);
    if (msgdelta>0 && (toc-currtime)>msgdelta)
      currtime = toc;
      [hrs,mins,secs]= secs2hms( (dictsize-j)*currtime/j );
      printf('updating: atom %d / %d, estimated remaining time: %02d:%02d:%05.2f', j, dictsize, hrs, mins, secs);
    end
  end
  if (msgdelta>0)
    printf('updating: atom %d / %d', j, dictsize);
  end
  
  Gamma = GammaT';
  
  
  %%%%%  compute error  %%%%%
  
  if (comperr)
    err(iter) = compute_err(D,Gamma,data);
  end
  if (testgen)
    GammaG = sparsecode(testdata,D,XtXg,G,thresh);
    gerr(iter) = compute_err(D,GammaG,testdata);
  end
   
  
  %%%%%  clear dictionary  %%%%%
  
  [D,cleared_atoms] = cleardict(D,Gamma,data,muthresh,unused_sigs,replaced_atoms);
  
  
  %%%%%  print info  %%%%%
  
  info = sprintf('Iteration %d / %d complete', iter, iternum);
  if (printerr)
    info = sprintf('%s, %s = %.4g', info, errstr, err(iter));
  end
  if (printgerr)
    info = sprintf('%s, testset %s = %.4g', info, errstr, gerr(iter));
  end
  if (printreplaced)
    info = sprintf('%s, replaced %d atoms', info, sum(replaced_atoms) + cleared_atoms);
  end
  
  if (printiter)
    disp(info);
  end
  
  if (savenum>0 && mod(iter,savenum)==0)
    save([filename , num2str(iter)]);
  end
  
end

if (savenum>=0)
  save([filename , num2str(iter)]);
end

return;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%            optimize_atom             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [atom,Gamma_j,unused_sigs,replaced_atoms] = optimize_atom(X,D,j,GammaT,unused_sigs,replaced_atoms)

global nonneg exactsvd

Gamma_j = GammaT(:,j);           % row in Gamma corresponding to current atom
data_indices = find(Gamma_j);    % data samples which use this atom

if (length(data_indices) < 1)
  E = sum((X-D*GammaT').^2);
  [d,i] = max(E(unused_sigs));
  atom = X(:,unused_sigs(i));
  atom = atom./norm(atom);
  Gamma_j = sparse(size(GammaT,1),1);
  unused_sigs = unused_sigs([1:i-1,i+1:end]);
  replaced_atoms(j) = 1;
  return;
end

smallGamma = GammaT(data_indices,:);
Dj = D(:,j);
gamma_j = Gamma_j(data_indices);

if (exactsvd)
  % The exact computation using svds
  [atom,s,gamma_j] = svds(X(:,data_indices)-D*smallGamma'+Dj*gamma_j',1);
  gamma_j = s*gamma_j;
  
  if (nonneg)
    atom(atom<0) = 0;
    atom = normcols(atom);
    
    % solve again for gamma_j using the updated atom
    gamma_j = rowlincomb(atom,X,1:size(X,1),data_indices)' - smallGamma*(atom'*D)' + (atom'*Dj)*gamma_j;
    gamma_j(gamma_j<0) = 0;
  end
else
  atom = collincomb(X,data_indices,gamma_j) - D*(gamma_j'*smallGamma)' + Dj*(gamma_j'*gamma_j);
  if (nonneg)
    atom(atom<0) = 0;
  end

  atom = normcols(atom);

  gamma_j = rowlincomb(atom,X,1:size(X,1),data_indices)' - smallGamma*(atom'*D)' + (atom'*Dj)*gamma_j;
  if (nonneg)
    gamma_j(gamma_j<0) = 0;
  end
end

Gamma_j(data_indices) = gamma_j;

return;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%             sparsecode               %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Gamma = sparsecode(data,D,XtX,G,thresh)

global CODE_SPARSITY codemode
global MEM_HIGH memusage
global nonneg ompfunc ompparams

if (memusage < MEM_HIGH)
  Gamma = ompfunc(D,data,G,thresh,ompparams{:});
  
else  % memusage is high
  
  if (codemode == CODE_SPARSITY)
    Gamma = ompfunc(D'*data,G,thresh,ompparams{:});
    
  else
    if (nonneg)
      Gamma = ompfunc(D,data,D'*data,G,thresh,ompparams{:});
    else
      Gamma = ompfunc(D'*data,XtX,G,thresh,ompparams{:});
    end
  end
  
end

return;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%             compute_err              %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function err = compute_err(D,Gamma,data)
  
global CODE_SPARSITY codemode

if (codemode == CODE_SPARSITY)
  err = sqrt(sum(sum((data-D*Gamma).^2))/numel(data));
else
  err = nnz(Gamma)/size(data,2);
end

return;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%           cleardict                  %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [D,cleared_atoms] = cleardict(D,Gamma,X,muthresh,unused_sigs,replaced_atoms)

use_thresh = 4;  % at least this number of samples must use the atom to be kept

dictsize = size(D,2);
E = sum((X-D*Gamma).^2);
G = D'*D;
G(diag_ids(G)) = 0;

cleared_atoms = 0;

for j = 1:dictsize
  if ( (max(abs(G(j,:))) > muthresh) || ...
       (~replaced_atoms(j) && length(find(abs(Gamma(j,:))>1e-7)) < use_thresh) )
    [y,i] = max(E(unused_sigs));
    D(:,j) = X(:,unused_sigs(i))/norm(X(:,unused_sigs(i)));
    unused_sigs = unused_sigs([1:i-1,i+1:end]);
    G_j = D'*D(:,j);
    G_j(j) = 0;
    G(:,j) = G_j;
    G(j,:) = G_j';
    cleared_atoms = cleared_atoms+1;
  end
end;

return;


