function [y,D] = ksvddenoise(params,msgdelta)
%KSVDDENOISE K-SVD denoising.
%  [Y,D] = KSVDDENOISE(PARAMS) denoises the specified (possibly
%  multi-dimensional) signal using the KSVD denoising algorithm. Y is the
%  denoised signal, and D is the trained dictionary.
%
%  [Y,D] = KSVDDENOISE(PARAMS,MSGDELTA) specifies whether the function
%  should print messages as it runs, and their level of detail. Setting
%  MSGDELTA to 0 cancels all messages, causing the function to run in
%  silent mode. When MSGDELTA is larger than zero, KSVD and OMPDENOISE
%  display messages every MSGDELTA seconds, and KSVD displays additional
%  information corresponding to the VERBOSE=3 setting. When MSGDELTA is
%  less than zero, KSVD and OMPDENOISE display messages every abs(MSGDELTA)
%  seconds, and KSVD displays additional information corresponding to the
%  VERBOSE=4 setting. The default value for MSGDELTA is 10.
%
%  Required fields in PARAMS:
%    x          -  the signal to denoise (can be multi-dimensional). x
%                  should be of type double, and, for psnr computations,
%                  its values are assumed to be in the range [0,1] (see
%                  parameter maxval below).
%    blocksize  -  the block size for the denoising. should be either an
%                  array of the form [N1 N2 ... Np], where p is the
%                  number of dimensions of x, or simply a scalar N, which
%                  is shorthand for the square block size [N N ... N].
%    dictsize   -  the number of dictionary elements to train.
%    trainnum   -  the number of blocks to use for the training step.
%    psnr       -  noise power in dB. used to determine the noise standard
%                  deviation and the target error for sparse coding each
%                  block.
%
%  Optional fields in PARAMS:
%    initdict   -  if present, specifies the initial dictionary for the
%                  training process. may either be an explicit matrix (of
%                  size prod(blocksize) X dictsize), or an array of length
%                  dictsize specifying the indices of the training blocks
%                  to be used as the initial atoms. when not present, the
%                  dictionary is initialized with randomly selected
%                  training blocks.
%    maxval     -  the maximal intensity value allowed in the signal (used
%                  for PSNR computations). by default, signals are assumed
%                  to contain values in the range [0,1]. if the maxval
%                  parameter is specified, the range is assumed to be
%                  [0,maxval]. see function PSNR for more information.
%    stepsize   -  the interval in voxels between neighboring signal
%                  blocks. by default, all overlapping signal blocks are
%                  denoised and averaged. when stepsize is set to the array
%                  [S1 S2 ... Sp] (where p is the number of dimensions of
%                  x), the distance between processed blocks in the i-th
%                  dimension becomes Si. stepsize can also be specified as
%                  a scalar S, shorthand for the step size [S S ... S].
%                  Each Si must be >= 1, and, to avoid cases where the
%                  boundary of x is not covered by any block, it is
%                  recommended that (size(X,i)-Ni) be a multiple of Si. the
%                  default behavior of the function corresponds to the
%                  setting stepsize==1.
%
%
%  Optional fields in PARAMS (advanced):
%    iternum    -  number of KSVD training iterations (default = 20).
%    gain       -  the noise gain. a positive value (usually close to 1)
%                  controling the target error when sparse coding each
%                  block. when gain=1, the target error is precisely the
%                  value derived from the psnr. when gain is different than
%                  1, the target error becomes the gain times the value
%                  derived from the psnr. default is gain = 1.15.
%    lambda     -  the relative weight given to the input (noisy) signal in
%                  determining the denoised output. for each output sample,
%                  covered by K overlapping blocks, its value is computed
%                  by weighing the K denoised samples against the original
%                  noisy sample. in the averaging process, each denoised
%                  sample is attributed a relative weight of 1, and the
%                  noisy sample is attributed a relative weight of lambda.
%                  this way, samples with less confidence (i.e. covered by
%                  less blocks) are regularized by giving larger weight to
%                  the original sample. default value is maxval/(10*sigma),
%                  where sigma is the standard deviation of the noise.
%    exact      -  specifies whether KSVD should use the approximate or
%                  exact atom update computation. by default, the
%                  approximate method is used, however when exact is
%                  specified and is non-zero, the exact method is used
%                  instead. see function KSVD for more information.
%    maxatoms   -  specifies the limit on the number of atoms used to code
%                  each signal block (see parameter 'maxatoms' in
%                  OMP2/NNOMP2). default value is prod(blocksize)/2, i.e.
%                  half the number of samples in a block.
%    memusage   -  used to control the memory requirements of the function.
%                  can be either 'high', 'normal' (default value) or 'low'.
%                  when memusage is specified, both KSVD and OMPDENOISE are
%                  executed using their corresponding memusage settings.
%                  additionally, when set to 'low', the use of the
%                  functions OMPDENOISE2 and OMPDENOISE3 to accelerate
%                  denoising of 2D and 3D signals is disabled, reverting to
%                  the function OMPDENOISE in any case.
%                  
%
%  References:
%  [1] M. Elad and M. Aharon, "Image Denoising Via Sparse and Redundant
%      representations over Learned Dictionaries", the IEEE Trans. on Image
%      Processing, Vol. 15, no. 12, pp. 3736-3745, December 2006.
%
%  See also KSVD, OMPDENOISE, OMP, OMP2.


%  Ron Rubinstein
%  Computer Science Department
%  Technion, Haifa 32000 Israel
%  ronrubin@cs
%
%  October 2007


%%%%% parse input parameters %%%%%

x = params.x;
p = ndims(x);
psnrval = params.psnr;
blocksize = params.blocksize;
trainnum = params.trainnum;


% blocksize %
if (numel(blocksize)==1)
  blocksize = ones(1,p)*blocksize;
end

% maxval %
if (isfield(params,'maxval'))
  maxval = params.maxval;
else
  maxval = 1;
end


% gain %
if (isfield(params,'gain'))
  gain = params.gain;
else
  gain = 1.15;
end


% verbose %
if (nargin<2)
  msgdelta = 10;
end

if (msgdelta == 0)
  verbose = 0;
elseif (msgdelta>0)
  verbose = 3;
else
  verbose = 4;
end
msgdelta = abs(msgdelta);


% remove fields in params which may potentially affect the KSVD function
if (isfield(params,'filename'))
  params = rmfield(params,'filename');
end
if (isfield(params,'savenum'))
  params = rmfield(params,'savenum');
end
if (isfield(params,'testdata'))
  params = rmfield(params,'testdata');
end


%%%% create training data %%%

ids = cell(p,1);
[ids{:}] = reggrid(size(x)-blocksize+1, trainnum, 'eqdist');
params.data = sampgrid(x,blocksize,ids{:});
params.data = remove_dc(params.data,'columns');


%%%%% KSVD training %%%%%

params.codemode = 'error';
sigma = psnr2std(psnrval,maxval);
params.Edata = sqrt(prod(blocksize)) * sigma * gain;   % target error for omp

if (msgdelta>0)
  disp('KSVD training...');
end
D = ksvd(params,verbose,msgdelta);


%%%%%  denoise the signal  %%%%%

if (~isfield(params,'lambda'))
  params.lambda = maxval/(10*sigma);
end

params.D = D;
if (msgdelta>0)
  disp('OMP denoising...');
end

if (ndims(x)>3  || (isfield(params,'memusage') && strcmpi(params.memusage,'low')))
  y = ompdenoise(params,msgdelta);
elseif (ndims(x)==2)
  y = ompdenoise2(params,msgdelta);
else
  y = ompdenoise3(params,msgdelta);
end

return;

