function y = ompdenoise(params,msgdelta)
%OMPDENOISE OMP denoising.
%  Y = OMPDENOISE(PARAMS) denoises a (possibly multi-dimensional) signal
%  using OMP denoising. OMPDENOISE denoises each block of the noisy signal
%  individually by first removing its mean and then performing OMP. The
%  denoised blocks are averaged when they overlap in the output signal.
%
%  Y = OMPDENOISE(PARAMS,MSGDELTA) specifies whether the function should
%  print messages as it runs. When MSGDELTA is 0, no messages are
%  displayed. When MSGDELTA is larger than zero, a message is displayed
%  approximately every MSGDELTA seconds.
%
%
%  Required fields in PARAMS:
%    x          -  the signal to denoise (can be multi-dimensional). x
%                  should be of type double, and (for psnr computations)
%                  its values are assumed to be in the range [0,1] (see
%                  parameter maxval below).
%    blocksize  -  the block size for the denoising. should be either an
%                  array of the form [N1 N2 ... Np], where p is the
%                  number of dimensions of x, or simply a scalar N, which
%                  is shorthand for the square block size [N N ... N].
%    D          -  the dictionary used to denoise each block in x
%    psnr       -  noise power in dB. used to determine the noise standard
%                  deviation and consequently the target error for sparse
%                  coding each block.
%
%
%  Optional fields in PARAMS:
%    maxval     -  the maximal intensity value allowed in the signal (used
%                  for PSNR computations). by default, signals are assumed
%                  to have values in the range [0,1]. if maxval is
%                  specified, the range is replaced by [0,maxval]. see
%                  function PSNR for more information.
%    stepsize   -  the interval in voxels between neighboring signal
%                  blocks. by default, all overlapping signal blocks are
%                  denoised and averaged. when stepsize is set to the array
%                  [S1 S2 ... Sp] (where p is the number of dimensions of
%                  x), the distance between processed blocks in the i-th
%                  dimension becomes Si. stepsize can also be specified as
%                  a scalar S, shorthand for the step size [S S ... S].
%                  Each Si must be >= 1, and, to avoid cases where the
%                  boundary of x is not covered by any block, it is
%                  recommended that (size(X,i)-Ni) be a multiple of Si. the
%                  default behavior of the function corresponds to the
%                  setting stepsize==1.
%
%
%  Optional fields in PARAMS (advanced):
%    gain       -  the noise gain. a positive value (usually close to 1)
%                  controling the target error when sparse coding each
%                  block. by default, gain=1 and the target error is
%                  precisely the value derived from the psnr. when gain is
%                  different from 1, the target error becomes the gain
%                  times the value derived from the psnr.
%    lambda     -  the relative weight given to the input (noisy) signal in
%                  determining the denoised output. for each output sample,
%                  covered by K overlapping blocks, its value is computed
%                  by weighing the K denoised samples against the original
%                  noisy sample. in the averaging process, each denoised
%                  sample is attributed a relative weight of 1, and the
%                  noisy sample is attributed a relative weight of lambda.
%                  this way, samples with less confidence (i.e. covered by
%                  less blocks) are regularized by giving larger weight to
%                  the original sample. by default, lambda=0, so each
%                  output sample is precisely the average of its denoised
%                  versions.
%    maxatoms   -  specifies the limit on the number of atoms used to code
%                  each signal block (see parameter 'maxatoms' in
%                  OMP2/NNOMP2). default value is prod(blocksize)/2, i.e.
%                  half the number of samples in a block.
%    memusage   -  this parameter can be used to control the memory
%                  requirements of the function. memusage can be either
%                  'high', 'normal' (default value) or 'low'. when set to
%                  'low', the matrix G=D'*D is not precomputed before
%                  calling OMP, thus reducing memory requirements when the
%                  dictionary is highly redundant. 'high' and 'normal' are
%                  equivalent settings. note that setting memusage to 'low'
%                  will increase runtime substantially.
%
%
%  References:
%  [1] M. Elad and M. Aharon, "Image Denoising Via Sparse and Redundant
%      representations over Learned Dictionaries", the IEEE Trans. on Image
%      Processing, Vol. 15, no. 12, pp. 3736-3745, December 2006.
%
%  See also OMPDENOISE2, OMPDENOISE3, KSVDDENOISE, OMP, OMP2.


%  Ron Rubinstein
%  Computer Science Department
%  Technion, Haifa 32000 Israel
%  ronrubin@cs
%
%  November 2007


% parse input arguments %

x = params.x;
p = ndims(x);
D = params.D;
psnrval = params.psnr;
blocksize = params.blocksize;

% blocksize %
if (numel(blocksize)==1)
  blocksize = ones(1,p)*blocksize;
end

% maxval %
if (isfield(params,'maxval'))
  maxval = params.maxval;
else
  maxval = 1;
end

% gain %
if (isfield(params,'gain'))
  gain = params.gain;
else
  gain = 1;
end

% lambda %
if (isfield(params,'lambda'))
  lambda = params.lambda;
else
  lambda = 0;
end

% maxatoms %
if (isfield(params,'maxatoms'))
  maxatoms = params.maxatoms;
else
  maxatoms = floor(prod(blocksize)/2);
end

% stepsize %
if (isfield(params,'stepsize'))
  stepsize = params.stepsize;
  if (numel(stepsize)==1)
    stepsize = ones(1,p)*stepsize;
  end
else
  stepsize = ones(1,p);
end
if (any(stepsize<1))
  error('Invalid step size.');
end


sigma = maxval / 10^(psnrval/20);                 % noise standard dviation
epsilon = sqrt(prod(blocksize)) * sigma * gain;   % target error for omp


MEM_LOW = 1;
MEM_NORMAL = 2;
MEM_HIGH = 3;

if (isfield(params,'memusage'))
  switch lower(params.memusage)
    case 'low'
      memusage = MEM_LOW;
    case 'normal'
      memusage = MEM_NORMAL;
    case 'high'
      memusage = MEM_HIGH;
    otherwise
      error('Invalid memory usage mode');
  end
else
  memusage = MEM_NORMAL;
end


% denoise the signal %

if (memusage >= MEM_NORMAL)
  G = D'*D;
else
  G = [];
end

y = zeros(size(x),class(x));

% ids{} contains the indices of the current block
ids = cell(p,1);
for j = 1:p
  ids{j} = 1:blocksize(j);
end

% lastids contains the indices of the last block in each dimension
lastids = stepsize .* floor((size(x)-blocksize)./stepsize) + 1;

blocknum = prod(floor((size(x)-blocksize)./stepsize) + 1);
tic; starttime = toc; currtime = starttime;
for i = 1 : blocknum
 
  block = x(ids{:});
  block = block(:);
  dc = mean(block);
  block = block-dc;
  y(ids{:}) = y(ids{:}) + reshape(D*omp2(D'*block,sum(block.*block),G,epsilon,'maxatoms',maxatoms),blocksize) + dc;

  % increment block ids
  if (i<blocknum)
    j = 1;
    while (ids{j}(1) == lastids(j))
      ids{j} = 1:blocksize(j);
      j = j+1;
    end
    ids{j} = ids{j}+stepsize(j);
  end
  
  % display status
  if (nargin>=2 && msgdelta>0 && (toc-currtime)>msgdelta)
    currtime = toc;
    [hrs,mins,secs]= secs2hms( (blocknum-i)*currtime/i );
    printf('ompdenoise: block %d / %d, estimated remaining time: %02d:%02d:%05.2f', i, blocknum, hrs, mins, secs);
  end

end

if (nargin>=2 && msgdelta>0)
  printf('ompdenoise: block %d / %d', i, blocknum);
end

% cnt is a matrix the same size as x, containing the number of times each
% voxel in x was covered by a block (for averaging purposes)

cnt = ones(size(x));
for k = 1:ndims(x)
  
  % this code is modified from function NDGRID, and used to compute one
  % output argument of NDGRID at a time (to conserve memory)
  ids = (1:size(x,k))';
  s = size(x); s(k) = [];
  ids = reshape(ids(:,ones(1,prod(s))),[length(ids) s]);
  ids = permute(ids,[2:k 1 k+1:ndims(x)]);
  
  cnt = cnt .* max( min(floor((ids-1)/stepsize(k)),floor((size(x,k)-blocksize(k))/stepsize(k))) - ...
                    max(ceil((ids-blocksize(k))/stepsize(k)),0) + 1 , 0 );
end

y = (y+lambda*x)./(cnt + lambda);

